require 'mature_factory/features/assemble/method_missing_decoration'
require 'mature_factory/features/assemble/id'
require 'mature_factory/features/assemble/proxy'
require 'mature_factory/features/assemble/abstract_builder'
require 'mature_factory/features/assemble/helpers'
require 'mature_factory/features/assemble/instantiation_error'
require 'mature_factory/features/assemble/error'
require 'mature_factory/features/assemble/exeptional'


module MatureFactory
  module Features
    module Assemble

      class Error < StandardError; end

      def self.create_init_error(e, id, init_attrs)
        new_error = InstantiationError.new(id, init_attrs)
        new_error.set_backtrace(e.backtrace)
        new_error
      end

      def self.create_error(e, id)
        new_error = Error.new(e.message, id)
        new_error.set_backtrace(e.backtrace)
        new_error
      end

      module ClassMethods
        class << self
          attr_accessor :__mf_assembler_name__
        end

        def self.included(receiver)
          receiver.extend self
        end

        def self.extended(receiver)
          receiver.produces :flatten_structs, :wrapped_structs, :nested_structs
          receiver.extend Helpers
        end

        def wrap(title, base_class: Class.new(Object), init: nil, delegate: false, on_exception: nil, &block)
          wrapped_struct(title, base_class: base_class.include(Exceptional), init: init)
          AbstractBuilder.new(self).wrap(title, delegate, on_exception, &block)
        end

        def nest(title, base_class: Class.new(Object), init: nil, delegate: false, on_exception: nil, &block)
          nested_struct(title, base_class: base_class.include(Exceptional), init: init)
          AbstractBuilder.new(self).nest(title, delegate, on_exception, &block)
        end

        def flat(title, base_class: Class.new(Object), init: nil, on_exception: nil, &block)
          flatten_struct(title, base_class: base_class.include(Exceptional), init: init)
          AbstractBuilder.new(self).flat(title, on_exception, &block)
        end

        def on_exception(&block)
          block_given? ? @on_exception = block : @on_exception
        end
      end

      module InheritanceHelpers
        def inherited(subclass)
          super
          subclass.on_exception(&self.on_exception)
        end
      end

      def self.included(receiver)
        receiver.extend ClassMethods
        receiver.extend InheritanceHelpers
        receiver.on_exception do |e, result|
          raise e
        end
      end
    end
  end
end
