require 'mature_factory/features/assemble/strategies/execute'
require 'mature_factory/features/assemble/strategies/enumerate'
require 'mature_factory/features/assemble/strategies/trace'
require 'mature_factory/features/assemble/strategies/reverse'
require 'mature_factory/features/assemble/strategies/prepare'
require 'mature_factory/features/assemble/strategies/transform'
require 'mature_factory/features/assemble/strategies/inject'
require 'mature_factory/features/assemble/strategies/init'
require 'mature_factory/features/assemble/strategies/link'
require 'mature_factory/features/assemble/strategies/mount'

module MatureFactory::Features::Assemble
  class AbstractBuilder < Struct.new(:factory)
    def self.link_with_delegation(target, object, accessor, delegate)
      target.define_singleton_method(:__mf_predecessor__) { accessor }
      target.define_singleton_method(accessor) { object }
      target.extend(MethodMissingDecoration) if delegate
    end

    def nest(name, delegate, on_exception, &block)
      factory.class_eval(&block)
      factory.define_singleton_method(:"build_#{name}") do |object: nil, title: nil, group: nil, &on_create|
        raise(ArgumentError, 'Please provide object: and title: arguments') if title.nil? ^ object.nil?
        trace = Strategies::Trace.new(&block)
        reverse = Strategies::Reverse.new(trace)
        prepare = Strategies::Prepare.new(reverse, self)
        pipe = Strategies::Transform.new(prepare, group, title, object).call
        list = pipe.map(&:last).map(&:to_sym)
        top = public_send(:"new_#{name}_nested_struct_instance")
        top.on_exception = on_exception
        inject = Strategies::Inject.new(list, &on_create)
        init = Strategies::Init.new(inject, pipe, top)
        link = Strategies::Link.new(init, delegate)
        enumerate = Strategies::Enumerate.new(link, top, self, delegate)
        Proxy.new(enumerate)
      end
    end

    def wrap(name, delegate, on_exception, &block)
      factory.class_eval(&block)
      factory.define_singleton_method(:"build_#{name}") do |object: nil, title: nil, group: nil, &on_create|
        raise(ArgumentError, 'Please provide object: and title: arguments') if (group.nil? && title.nil?) ^ object.nil?
        trace = Strategies::Trace.new(&block)
        prepare = Strategies::Prepare.new(trace, self)
        pipe = Strategies::Transform.new(prepare, group, title, object).call
        list = pipe.map(&:last).map(&:to_sym)
        top = public_send(:"new_#{name}_wrapped_struct_instance")
        top.on_exception = on_exception
        inject = Strategies::Inject.new(list, &on_create)
        init = Strategies::Init.new(inject, pipe, top)
        link = Strategies::Link.new(init, delegate)
        enumerate = Strategies::Enumerate.new(link, top, self, delegate)
        Proxy.new(enumerate)
      end
    end

    def flat(name, on_exception, &block)
      factory.class_eval(&block)
      factory.define_singleton_method(:"build_#{name}") do |object: nil, title: nil, group: nil, &on_create|
        raise(ArgumentError, 'Please provide object: and title: arguments') if (group.nil? && title.nil?) ^ object.nil?
        trace = Strategies::Trace.new(&block)
        prepare = Strategies::Prepare.new(trace, self)
        list = prepare.call.to_a
        list.unshift([group, title, proc{object}]) unless object.nil?
        list = list.transpose
        ids = list[0..1].transpose.map{|g,e| MatureFactory::Features::Assemble::Id.new(g,e)}
        inject = Strategies::Inject.new(ids.map(&:to_sym), &on_create)
        top = public_send(:"new_#{name}_flatten_struct_instance")
        top.on_exception = on_exception
        mount = Strategies::Mount.new(inject, ids.zip(list[-1]), top, title)
        enumerate = Strategies::Execute.new(mount, top, self)
        Proxy.new(enumerate)
      end
    end
  end
end
