module MatureFactory::Features::Assemble::Strategies
  class Execute < Struct.new(:strategy, :top, :factory, :delegate)
    def call(&block)
      catch(:halt) do
        begin
          block_given? ? build_plan.each(&block) : build_plan.to_a
        rescue => e
          handle_exception(e)
        end
      end
      prepare_top
    end

    private

    def build_plan
      @build_plan ||= strategy.call
    end

    def handle_exception(e)
      top.on_exception ||= factory.on_exception
      top.exception = wrap_error(e)
      exetute_on_exception_strategy(top.exception, top.on_exception, top)
    end

    def wrap_error(e)
      return e if e.is_a?(MatureFactory::Features::Assemble::InstantiationError)
      MatureFactory::Features::Assemble.create_error(e, strategy.step_id)
    end

    def exetute_on_exception_strategy(e, strategy, actor)
      case strategy
      when Proc
        strategy.call(e, top)
      when :halt
        throw :halt
      when :raise
        raise(e)
      else
        nil
      end
    end

    def prepare_top
      top
    end
  end
end
