module MatureFactory::Features::Assemble::Strategies
  class Mount < Struct.new(:strategy, :pipe, :top, :step)
    attr_reader :object, :step_id

    def call
      Enumerator.new do |yielder|
        init_params = strategy.call
        pipe.each do |id, init_proc|
          @step_id = id
          init_attrs = Array(init_params[id.to_sym])
          begin
            layer_object = init_proc.call(*init_attrs)
          rescue => e
            next unless handle_exception(e, id, init_attrs)
          end
          top.define_singleton_method(id.title.to_sym) { layer_object }
          yielder << [layer_object, id]
        end
      end
    end

    def handle_exception(e, id, init_attrs)
      exception = MatureFactory::Features::Assemble.create_init_error(e, id, init_attrs)
      top.exception = exception
      raise exception unless top.ignore_exceptions?
    end
  end
end
